"""Tests for the gillcup.animations module"""

from __future__ import unicode_literals, division, print_function

import math
import weakref
import gc

from pytest import raises, mark

from gillcup import (
    Clock, AnimatedProperty, TupleProperty, Animation, animation, effect)
from gillcup import easing as easing_mod


class ToneWithProperties(object):
    """Test class with some animated properties"""
    pitch = AnimatedProperty(440)
    volume = AnimatedProperty(0)
    x, y, z = position = TupleProperty(0, 0, 0)


def pytest_generate_tests(metafunc):
    """Let Tone be injected in all tests!"""
    if "Tone" in metafunc.funcargnames:
        metafunc.parametrize("Tone", [ToneWithProperties])


def test_animation(Tone):
    """Test a simple animation"""
    clock = Clock()
    tone = Tone()
    clock.schedule(Animation(tone, 'pitch', 450, time=5))
    assert tone.pitch == 440
    clock.advance(1)
    assert tone.pitch == 442
    clock.advance(3)
    assert tone.pitch == 448
    clock.advance(1)
    assert tone.pitch == 450
    clock.advance(1)
    assert tone.pitch == 450
    tone.pitch = 440
    assert tone.pitch == 440
    clock.schedule(Animation(tone, 'pitch', 480, time=4))
    clock.advance(1)
    assert tone.pitch == 450
    clock.advance(1)
    assert tone.pitch == 460
    tone.pitch = 500
    assert tone.pitch == 500
    del tone.pitch
    assert tone.pitch == 440

    with raises(ValueError):
        Animation(tone, 'pitch', 450, target=460)


def test_animation_scheduling(Tone):
    """Test the scheduling of animations"""
    clock = Clock()
    tone = Tone()
    tone.volume = 60
    action = Animation(tone, 'pitch', 450, time=5)
    action += Animation(tone, 'volume', 0, time=2)
    clock.schedule(action)
    assert tone.pitch == 440 and tone.volume == 60
    clock.advance(1)
    assert tone.pitch == 442 and tone.volume == 60
    clock.advance(3)
    assert tone.pitch == 448 and tone.volume == 60
    clock.advance(1)
    assert tone.pitch == 450 and tone.volume == 60
    clock.advance(1)
    assert tone.pitch == 450 and tone.volume == 30
    clock.advance(1)
    assert tone.pitch == 450 and tone.volume == 0
    clock.advance(1)
    assert tone.pitch == 450 and tone.volume == 0


def test_tuple_property(Tone):
    """Test animating a TupleProperty"""
    clock = Clock()
    tone = Tone()
    action = Animation(tone, 'x', 2, time=2)
    action |= Animation(tone, 'y', 4, time=2)
    action |= Animation(tone, 'z', 6, time=2)
    action += Animation(tone, 'position', 0, 0, 0, time=2)
    clock.schedule(action)
    assert (tone.x, tone.y, tone.z) == tone.position == (0, 0, 0)
    clock.advance(1)
    assert (tone.x, tone.y, tone.z) == tone.position == (1, 2, 3)
    clock.advance(1)
    assert (tone.x, tone.y, tone.z) == tone.position == (2, 4, 6)
    clock.advance(1)
    assert (tone.x, tone.y, tone.z) == tone.position == (1, 2, 3)
    clock.advance(1)
    assert (tone.x, tone.y, tone.z) == tone.position == (0, 0, 0)
    tone.position = 1, 2, 0
    assert (tone.x, tone.y, tone.z) == tone.position == (1, 2, 0)
    tone.z = 3
    assert (tone.x, tone.y, tone.z) == tone.position == (1, 2, 3)


def test_target_animation(Tone):
    """Test animating the targer of an animation... Meta!"""
    clock = Clock()
    tone = Tone()
    action = Animation(tone, 'volume', 2, time=2, dynamic=True)
    action.chain(Animation(action, 'target', 0, time=2))
    clock.schedule(action)
    clock.advance(1)
    assert tone.volume == 1
    clock.advance(1)
    assert tone.volume == 2
    clock.advance(1)
    assert tone.volume == 1
    clock.advance(1)
    assert tone.volume == 0


def test_tuple_target_animation(Tone):
    """Test animating the targer of a TupleAnimation"""
    clock = Clock()
    tone = Tone()
    action = Animation(tone, 'position', 2, 4, 6, time=2, dynamic=True)
    action.chain(Animation(action, 'target', -2, -4, -6, time=2))
    clock.schedule(action)
    clock.advance(1)
    assert tone.position == (1, 2, 3)
    clock.advance(1)
    assert tone.position == (2, 4, 6)
    clock.advance(1)
    assert tone.position == (0, 0, 0)
    clock.advance(1)
    assert tone.position == (-2, -4, -6)


def test_additive_animation(Tone):
    """Test the animation.Add flavor of animation"""
    clock = Clock()
    tone = Tone()
    base_action = Animation(tone, 'pitch', 420, delay=3, time=2)
    action = animation.Add(tone, 'pitch', 20, time=2)
    clock.schedule(base_action)
    clock.schedule(action)
    clock.advance(1)
    assert tone.pitch == 450
    clock.advance(1)
    assert tone.pitch == 460
    clock.advance(1)
    assert tone.pitch == 460
    clock.advance(1)
    assert tone.pitch == 450
    clock.advance(1)
    assert tone.pitch == 440


def test_multiplicative_animation(Tone):
    """Test the animation.Multiply flavor of animation"""
    clock = Clock()
    tone = Tone()
    base_action = Animation(tone, 'pitch', 220, time=2, delay=3)
    action = animation.Multiply(tone, 'pitch', 2, time=2)
    clock.schedule(base_action)
    clock.schedule(action)
    clock.advance(1)
    assert tone.pitch == 440 * 1.5
    clock.advance(1)
    assert tone.pitch == 440 * 2
    clock.advance(1)
    assert tone.pitch == 440 * 2
    clock.advance(1)
    assert tone.pitch == 330 * 2
    clock.advance(1)
    assert tone.pitch == 220 * 2


def test_computed_animation(Tone):
    """Test the animation.Computed flavor of animation"""
    def _compute(t):
        return math.sin(t) + 440
    clock = Clock()
    tone = Tone()
    action = animation.Computed(tone, 'pitch', func=_compute, time=2)
    clock.schedule(action)
    clock.advance(0)
    assert tone.pitch == 440
    clock.advance(1)
    assert tone.pitch == 440 + math.sin(0.5)
    clock.advance(1)
    assert tone.pitch == 440 + math.sin(1)
    clock.advance(1)
    assert tone.pitch == 440 + math.sin(1)


def test_delay(Tone):
    """Test the delay argument"""
    clock = Clock()
    tone = Tone()
    action = Animation(tone, 'pitch', 450, time=2, delay=1)
    clock.schedule(action)
    clock.advance(1)
    assert tone.pitch == 440
    clock.advance(1)
    assert tone.pitch == 445
    clock.advance(1)
    assert tone.pitch == 450
    clock.advance(1)
    assert tone.pitch == 450


def test_infinite_timing(Tone):
    """Test timing='infinite'"""
    clock = Clock()
    tone = Tone()
    action = Animation(tone, 'pitch', 450, time=1, timing='infinite')
    clock.schedule(action)
    assert tone.pitch == 440
    clock.advance(1)
    assert tone.pitch == 450
    clock.advance(1)
    assert tone.pitch == 460
    clock.advance(1)
    assert tone.pitch == 470
    clock.advance(0.5)
    assert tone.pitch == 475


def test_absolute_timing(Tone):
    """Test timing='absolute'"""
    clock = Clock()
    tone = Tone()
    clock.advance(1)
    action = Animation(tone, 'pitch', 450, time=1, timing='absolute')
    clock.schedule(action)
    clock.advance(0)
    assert tone.pitch == 450
    clock.advance(1)
    assert tone.pitch == 460
    clock.advance(1)
    assert tone.pitch == 470
    clock.advance(0.5)
    assert tone.pitch == 475


def test_custom_timing(Tone):
    """Test custom timing"""
    clock = Clock()
    tone = Tone()
    clock.advance(1)
    action = Animation(
        tone, 'pitch', 450, time=1, timing=lambda t, s, d: (t - s) / d / 2)
    clock.schedule(action)
    clock.advance(0)
    assert tone.pitch == 440
    clock.advance(1)
    assert tone.pitch == 445
    clock.advance(1)
    assert tone.pitch == 450
    clock.advance(1)
    assert tone.pitch == 455


def test_easing(Tone):
    """Test all kinds of easing functions"""
    easings = []
    for easing in [
            'linear', 'quadratic', 'cubic', 'quartic', 'quintic',
            'sine', 'exponential', 'circular']:
        easings.append(easing)
        for mod in [''] + '.in_ .out .in_out .out_in'.split():
            easings.append(easing + mod)
        easings.append(getattr(easing_mod, easing))
    for easing in (
            easing_mod.linear, easing_mod.exponential,
            easing_mod.bounce(2)):
        easings.append(easing)
        for mod in 'in_ out in_out out_in'.split():
            easings.append(getattr(easing, mod))
    for easing in easings:
        print(easing)
        clock = Clock()
        tone = Tone()
        action = Animation(tone, 'pitch', 450, time=2, easing=easing)
        clock.schedule(action)
        for dummy in range(21):
            clock.advance(0.05)
            assert 440 < tone.pitch < 450
        clock.advance(1)
        assert tone.pitch == 450
    for parameter in 0.5, 1, 2, 3:
        for easing in easings + [
                easing_mod.bounce(parameter),
                easing_mod.overshoot(parameter),
                easing_mod.elastic(parameter)]:
            print(easing)
            clock = Clock()
            tone = Tone()
            action = Animation(tone, 'pitch', 450, time=1, easing=easing)
            clock.schedule(action)
            assert tone.pitch == 440
            clock.advance(1)
            assert tone.pitch == 450
    assert easing_mod.circular(2) == 1  # possible sqrt of neg. number


def test_easing_normalization():
    """Test easingfunc normalization works as advertised"""
    for base_func in (lambda t: t * 4 + 3), (lambda t: t):
        f = easing_mod.normalized(base_func)
        assert [t / 16 for t in range(20)] == [f(t / 16) for t in range(20)]


def test_simple_resource_freeing(Tone):
    """Assert unneeded animations don't linger around"""
    clock = Clock()
    tone = Tone()
    anim = Animation(tone, 'pitch', 450, time=1)
    clock.schedule(anim)
    ref = weakref.ref(anim)
    del anim
    clock.advance(1)
    gc.collect()
    assert tone.pitch == 450
    assert ref() is None


def test_resource_freeing_of_parent(Tone):
    """Assert unneeded animations don't linger around if shadowed"""
    clock = Clock()
    tone = Tone()
    anim = Animation(tone, 'pitch', 450, time=1)
    clock.schedule(anim)
    clock.schedule(Animation(tone, 'pitch', 460, time=1))
    ref = weakref.ref(anim)
    del anim
    clock.advance(1)
    gc.collect()
    assert tone.pitch == 460
    assert ref() is None


def test_resource_freeing_of_add_parent(Tone):
    """Assert unneeded animations don't linger around if shadowed by Add"""
    clock = Clock()
    tone = Tone()
    anim = Animation(tone, 'pitch', 450, time=1)
    clock.schedule(anim)
    clock.schedule(animation.Add(tone, 'pitch', 50, time=10))
    ref = weakref.ref(anim)
    del anim
    clock.advance(1)
    gc.collect()
    assert tone.pitch == 455
    assert ref() is None


def test_resource_freeing_of_tuple_property_parent(Tone):
    """Assert unneeded animations don't linger around if shadowed by Add"""
    clock = Clock()
    tone = Tone()
    anim = Animation(tone, 'x', 10, time=1)
    clock.schedule(anim)
    ref = weakref.ref(anim)
    del anim
    clock.advance(1)
    gc.collect()
    assert tone.position == (10, 0, 0)
    assert ref() is None


@mark.parametrize('_base_param', [('position', 3), ('x', 1)])
@mark.parametrize('_add_param', [('position', 3), ('x', 1)])
def test_resource_freeing_of_tuple_add_parent(Tone, _base_param, _add_param):
    """Assert unneeded animations don't linger around if shadowed by Add

    Test with all combinations of whole tuple and single element
    """
    clock = Clock()
    tone = Tone()
    base_attr, base_num = _base_param
    add_attr, add_num = _add_param

    anim = Animation(tone, base_attr, *[10] * base_num, time=1)
    clock.schedule(anim)
    clock.schedule(animation.Add(tone, add_attr, *[100] * add_num, time=10))
    ref = weakref.ref(anim)
    del anim
    clock.advance(1)
    gc.collect()
    assert tone.x == 20
    assert ref() is None


def test_resource_freeing_of_tuple_property_set(Tone):
    """Assert unneeded effects don't linger around on direct tuple item set"""
    tone = Tone()
    #import pytest; pytest.set_trace()
    tone.x = 10
    gc.collect()
    assert tone.position == (10, 0, 0)
    assert isinstance(Tone.position.get_effect(tone), effect.ConstantEffect)


def test_resource_freeing_of_tuple_property_delete(Tone):
    """Assert unneeded effects don't linger around on direct tuple item set"""
    tone = Tone()
    #import pytest; pytest.set_trace()
    tone.position = 10, 10, 10
    del tone.x
    gc.collect()
    assert tone.position == (0, 10, 10)
    assert isinstance(Tone.position.get_effect(tone), effect.ConstantEffect)


def test_elementwise_tuple_animation(Tone):
    """Test animating each element of a tuple independently"""
    tone = Tone()
    clock = Clock()
    clock.schedule(Animation(tone, 'x', 10, time=10))
    clock.schedule(Animation(tone, 'y', 20, time=10))
    clock.schedule(Animation(tone, 'z', 30, time=10))
    assert tone.position == (tone.x, tone.y, tone.z) == (0, 0, 0)
    clock.advance(1)
    assert tone.position == (tone.x, tone.y, tone.z) == (1, 2, 3)
    clock.schedule(Animation(tone, 'x', 0, time=8))
    clock.schedule(Animation(tone, 'y', 0, time=8))
    clock.schedule(Animation(tone, 'z', 0, time=8))
    clock.advance(4)
    # Both animations are halfway through
    assert tone.position == (tone.x, tone.y, tone.z) == (5 / 2, 10 / 2, 15 / 2)
    clock.advance(4)
    assert tone.position == (tone.x, tone.y, tone.z) == (0, 0, 0)


def test_effect_apply(Tone):
    """Test direct application of an Effect"""
    tone = Tone()
    assert tone.pitch == 440
    effect.ConstantEffect(450).apply_to(tone, 'pitch')
    assert tone.pitch == 450



"""Tests for the gillcup.properties module"""

from pytest import raises

from gillcup import Clock, Animation
from gillcup.properties import VectorProperty, ScaleProperty


class Box(object):
    """A Test class with some properties"""
    position = VectorProperty(3)
    size = ScaleProperty(3)


def test_defaults():
    """Test ScaleProperty and VectorProperty defaults"""
    box = Box()
    assert box.position == (0, 0, 0)
    assert box.size == (1, 1, 1)


def test_vector_assignment():
    """Test assigning to a VectorProperty"""
    box = Box()
    box.position = 1,
    assert box.position == (1, 0, 0)
    box.position = 1, 2
    assert box.position == (1, 2, 0)
    box.position = 1,
    assert box.position == (1, 0, 0)
    box.position = 1, 2, 3
    assert box.position == (1, 2, 3)
    with raises(TypeError):
        box.position = 4
    with raises(ValueError):
        box.position = 1, 2, 3, 4


def test_scale_assignment():
    """Test assigning to a ScaleProperty"""
    box = Box()
    box.size = 2,
    assert box.size == (2, 2, 2)
    box.size = 2, 3
    assert box.size == (2, 3, 1)
    box.size = 2,
    assert box.size == (2, 2, 2)
    box.size = 2, 3, 4
    assert box.size == (2, 3, 4)
    box.size = 5
    assert box.size == (5, 5, 5)


def test_vector_animation():
    """Test animating a VectorProperty"""
    box = Box()
    clock = Clock()
    clock.schedule(Animation(box, 'position', 1))
    clock.advance(1)
    assert box.position == (1, 0, 0)
    clock.schedule(Animation(box, 'position', 1, 2))
    clock.advance(1)
    assert box.position == (1, 2, 0)
    clock.schedule(Animation(box, 'position', 1))
    clock.advance(1)
    assert box.position == (1, 0, 0)
    clock.schedule(Animation(box, 'position', 1, 2, 3))
    clock.advance(1)
    assert box.position == (1, 2, 3)
    with raises(ValueError):
        Animation(box, 'position', 1, 2, 3, 4)


def test_scale_animation():
    """Test animating a ScaleProperty"""
    box = Box()
    clock = Clock()
    clock.schedule(Animation(box, 'size', 2))
    clock.advance(1)
    assert box.size == (2, 2, 2)
    clock.schedule(Animation(box, 'size', 2, 3))
    clock.advance(1)
    assert box.size == (2, 3, 1)
    clock.schedule(Animation(box, 'size', 2))
    clock.advance(1)
    assert box.size == (2, 2, 2)
    clock.schedule(Animation(box, 'size', 2, 3, 4))
    clock.advance(1)
    assert box.size == (2, 3, 4)
    with raises(ValueError):
        Animation(box, 'size', 1, 2, 3, 4)


def test_no_error_in_do_replacements():
    """Test no error occurs when calling do_replacements on bogus object"""
    Box.size.do_replacements(None)



# Encoding: UTF-8
"""Gillcup's Animation classes

Animations are :mod:`Actions <gillcup.actions>` that modify
:mod:`animated properties <gillcup.properties>`.
To use one, create it and schedule it on a Clock.
Once an animation is in effect, it will smoothly change a property's value
over a specified time interval.

The value is computed as a tween between the property's original value and
the Animation's **target** value.
The tween parameters can be set by the **timing** and **easing** keyword
arguments.

The “original value” of a property is not fixed: it is whatever the
value would have been if this animation wasn't applied (in other words,
it's determined by the :mod:`~gillcup.effect` that was originally on the
property).
Also, if you set the **dynamic** argument to Animation, the animation's
*target* becomes an :class:`~gillcup.AnimatedProperty`.
Animating these allows one to create very complex effects in a modular way.
"""

from __future__ import unicode_literals, division, print_function

from six import string_types

from gillcup.actions import Action
from gillcup.effect import Effect, ConstantEffect
from gillcup.properties import AnimatedProperty
from gillcup import easing as easing_module


class Animation(Effect, Action):
    """An object that modifies an AnimatedProperty based on Clock time

    Positional init arguments:

    :argument instance:

        The object whose property is animated

    :argument property_name:

        Name of the animated property

    :argument target:

        Value at which the animation should arrive (tuple properties
        accept more arguments, i.e. ``Animation(obj, 'position', 1, 2, 3)``)

    Keyword init arguments:

    :argument time:

        The duration of the animation

    :argument delay:

        Delay between the time the animation is scheduled and its actual start

    :argument timing:

        A function that maps global time to animation's time.

        Possible values:

        *   ``None``: normalizes time so that 0 corresponds to the start of the
            animation, and 1 to the end (i.e. start + `time`); clamps to [0, 1]
        *   ``'infinite'``: same as above, but doesn't clamp: the animation
            goes forever on (in both directions; it only starts to take effect
            when it's scheduled, but a `delay` can cause negative local times).
            The animation's time is normalized to 0 at the start and
            1 at start + `time`.
        *   ``'absolute'``: the animation is infinite, with the same speed as
            with the 'infinite' option, but zero corresponds to the clock's
            zero.
            Useful for synchronized periodic animations.
        * `function(time, start, duration)`: apply a custom function

    :argument easing:

        An easing function to use. Can be either a one-argument
        function, or a dotted name which is looked up in the
        :mod:`gillcup.easing` module.

    :argument dynamic:

        If true, the **target** atribute becomes an AnimatedProperty, allowing
        for more complex animations.

    .. note::

        In order to conserve resources, ordinary Animations are released
        (replaced by a simple :class:`~gillcup.ConstantEffect`) when
        they are “done”.
        Arguments such as ``timing``, or the
        :class:`~gillcup.animation.Add` or :class:`~gillcup.animation.Multiply`
        animation subclasses, which allow the value to be modified after the
        ``time`` elapses, turn this behavior off by setting the ``dynamic``
        attibute to true.

        When subclassing Animation, remember to do the same if your subclass
        needs to change its value after ``time`` elapses.
        This includes cases where the value depends on the value of the
        previous (parent) animation.
    """
    dynamic = False

    start_time = None
    parent = None
    strength = 1

    def __init__(self, instance, property_name, *target, **kwargs):
        super(Animation, self).__init__()
        self.instance = instance
        self.property = self.get_property(instance, property_name)

        try:
            new_target = kwargs.pop('target')
        except KeyError:
            pass
        else:
            if target:
                raise ValueError(
                    'Target specified as both positional and keyword argument')
            target = new_target

        self.target = self.property.adjust_value(target)

        self.dynamic = (
            self.dynamic or 'timing' in kwargs or kwargs.pop('dynamic', False))
        if not self.dynamic:
            self.chain(lambda: self.property.do_replacements(instance))

        self.time = kwargs.pop('time', 1)
        self.delay = kwargs.pop('delay', 0)
        easing = kwargs.pop('easing', 'linear')
        timing = kwargs.pop('timing', None)

        if timing == 'infinite':
            self.get_time = self._infinite_timing
        elif timing == 'absolute':
            self.get_time = self._absolute_timing
        elif timing:
            self.get_time = lambda: timing(
                self.clock.time, self.start_time, self.time)

        if isinstance(easing, string_types):
            e = easing_module
            for attr in easing.split('.'):
                e = getattr(e, attr)
            self.easing = e
        else:
            self.easing = easing

    @classmethod
    def get_property(cls, instance, property_name):
        """Get a property object off an instance's class"""
        return getattr(type(instance), property_name)

    def __new__(cls, instance, property_name, *args, **kwargs):
        if kwargs.get('dynamic', False):
            # We need the target to act the same as the animated property
            # (wrt adjust_value: being scalar/tuple, etc).
            # An AnimatedProperty needs to be on a class, we can't just put
            # a descriptor on an instance.
            # So, we create a trivial subclass that has the target property.
            prop = cls.get_property(instance, property_name)

            class AnimatedAnimation(cls):
                """A more dynamic flavor of gillcup.Animation"""
                target = prop.get_target_property()
                strength = AnimatedProperty(1)
            ani_class = AnimatedAnimation
        else:
            ani_class = cls
        super_new = super(Animation, cls).__new__
        return super_new(ani_class, instance, property_name, *args, **kwargs)

    def __call__(self):
        self.expire()
        self.parent = self.property.animate(self.instance, self)
        self.start_time = self.clock.time + self.delay
        self.clock.schedule(self.trigger_chain, self.time + self.delay)

    @property
    def value(self):
        """Value to be used for the property this animation is on"""
        parent_value = self.parent.value
        target = self.target
        return self.property.tween_values(
            self.compute_value, parent_value, target)

    def get_time(self):  # pylint: disable=E0202
        """Get the local time for tweening, usually in the [0..1] range"""
        elapsed = self.clock.time - self.start_time
        if elapsed <= 0:
            return 0
        if elapsed >= self.time:
            return 1
        else:
            return elapsed / self.time
    get_time.finite = True

    def _absolute_timing(self):
        return self.clock.time / self.time

    def _infinite_timing(self):
        return (self.clock.time - self.start_time) / self.time

    def compute_value(self, previous, target):
        """Given the previous value and a target, compute value"""
        t = self.easing(self.get_time()) * self.strength
        return previous * (1 - t) + target * t

    def get_replacement(self):
        if not self.dynamic and self.get_time() >= 1:
            # Not gonna change from now on
            return ConstantEffect(self.value)
        else:
            self.parent = self.parent.get_replacement()
            return self


class Add(Animation):
    """An additive animation: the target value is added to the original
    """
    dynamic = True

    def compute_value(self, previous, target):
        t = self.easing(self.get_time())
        return previous + target * t


class Multiply(Animation):
    """A multiplicative animation: target value is multiplied to the original
    """
    dynamic = True

    def compute_value(self, previous, target):
        t = self.easing(self.get_time())
        return previous * ((1 - t) + target * t)


class Computed(Animation):
    """A custom-valued animation: the target is computed by a function

    Pass a **func** keyword argument with the function to the constructor.

    The function will get one argument: the time elapsed, normalized by the
    animation's `timing` function.
    """
    def __init__(self, instance, property_name, func, **kwargs):
        self.func = func
        prop = self.get_property(instance, property_name)
        kwargs.setdefault('target', [prop.default])
        super(Computed, self).__init__(instance, property_name, **kwargs)

    def compute_value(self, previous, target):
        t = self.get_time()
        return self.func(t)



"""Tests for the gillcup.clock module"""

from __future__ import unicode_literals, division, print_function

from pytest import raises

from gillcup import Clock, Subclock, Animation


def dummy_function():
    """Helper function; does nothing"""
    pass


def append_const(lst, value):
    """Helper callback; returns func that appends a constant to list"""
    def _append():
        lst.append(value)
    return _append


def append_time(lst, clock):
    """Helper callback; returns func that appends clock's time to list"""
    def _append():
        lst.append(clock.time)
    return _append


def test_clock():
    """Test that simple clock advancement works"""
    clock = Clock()
    assert clock.time == 0
    clock.advance(1)
    assert clock.time == 1


def test_scheduling():
    """Test that simple scheduling works"""
    clock = Clock()
    lst = []
    clock.schedule(append_const(lst, 'a'), 1)
    clock.schedule(append_const(lst, 'd'), 3)
    clock.schedule(append_const(lst, 'b'), 1)
    clock.schedule(append_const(lst, 'c'), 1)
    clock.advance(0)
    assert lst == []
    clock.advance(1)
    assert lst == ['a', 'b', 'c']
    clock.advance(1)
    assert lst == ['a', 'b', 'c']
    clock.advance(1)
    assert lst == ['a', 'b', 'c', 'd']


def test_speed():
    """Test that the speed property works"""
    clock = Clock()
    lst = []
    clock.schedule(append_const(lst, 'a'), 10)
    clock.schedule(append_const(lst, 'd'), 30)
    clock.schedule(append_const(lst, 'b'), 10)
    clock.schedule(append_const(lst, 'c'), 10)
    clock.speed = 10
    clock.advance(0)
    assert lst == []
    clock.advance(1)
    assert lst == ['a', 'b', 'c']
    clock.advance(1)
    assert lst == ['a', 'b', 'c']
    clock.advance(1)
    assert lst == ['a', 'b', 'c', 'd']
    clock.speed = -1
    with raises(ValueError):
        clock.advance(4)


def test_negative_dt():
    """Test that negative advancements are blocked"""
    clock = Clock()
    with raises(ValueError):
        clock.advance(-1)
    with raises(ValueError):
        clock.schedule(dummy_function, -1)


def test_zero_dt():
    """Test that a zero advancement triggers immediate events"""
    clock = Clock()
    lst = []
    clock.schedule(append_const(lst, 'a'), 0)
    clock.advance(0)
    assert lst == ['a']


def test_integer_times():
    """Test that we keep to int arithmetic as much as we can

    Ensure that even if the clock advances by float ticks, events scheduled
    at integer times get fired with clock.time being an integer.
    """
    clock = Clock()
    lst = []
    clock.schedule(append_time(lst, clock), 1)
    clock.schedule(append_time(lst, clock), 2)
    clock.schedule(append_time(lst, clock), 3)
    for dummy in range(30):
        clock.advance(0.3)
    assert lst == [1, 2, 3]
    for time in lst:
        assert type(time) == int


def test_update_function():
    """Test that schedule_update_function() works"""
    time_list = [0]
    clock = Clock()

    def _update_function():
        time_list[0] = clock.time

    def _do_assert():
        assert time_list[0] == clock.time

    clock.schedule_update_function(_update_function)
    _do_assert()
    clock.advance(1)
    _do_assert()
    clock.schedule(_do_assert, 1)
    clock.advance(2)
    _do_assert()


def test_unschedule_update_function():
    """Test that unschedule_update_function() works"""
    clock = Clock()

    def _update_function():
        raise AssertionError('Update function called!')

    clock.schedule_update_function(_update_function)
    clock.unschedule_update_function(_update_function)
    clock.advance(1)


def test_recursive_advance():
    """Test that calling advance() from within advance() is blocked"""
    clock = Clock()
    clock.schedule(lambda: clock.advance(1))
    with raises(RuntimeError):
        clock.advance(0)


def test_subclock():
    """Test Sublock works"""
    clock = Clock()
    subclock = Subclock(clock)
    subclock.advance(1)
    assert clock.time == 0
    assert subclock.time == 1
    clock.advance(1)
    assert clock.time == 1
    assert subclock.time == 2
    subclock.speed = 2
    clock.advance(1)
    assert clock.time == 2
    assert subclock.time == 4

    lst = []

    def _append_const(lst, value):
        def _append():
            print(value, clock.time, subclock.time)
            lst.append(value)
        return _append

    subclock.schedule(_append_const(lst, 'a'), 1)
    clock.schedule(_append_const(lst, 'b'), 1)
    subclock.schedule(_append_const(lst, 'c'), 2)
    clock.advance(1)
    assert lst == ['a', 'b', 'c']


def test_subclock_speedup():
    """Test Sublock.speed works when set as attribute"""
    clock = Clock()
    subclock = Subclock(clock)
    assert subclock.time == 0
    clock.advance(1)
    assert subclock.time == 1
    subclock.speed = 2
    clock.advance(1)
    assert subclock.time == 3


def test_subclock_speed_arg():
    """Test Sublock.speed works when passed as argument"""
    clock = Clock()
    subclock = Subclock(clock, speed=2)
    assert subclock.time == 0
    clock.advance(1)
    assert subclock.time == 2


def test_subclock_animated_speed():
    """Test behavior is consistent when Sublock.speed is animated"""
    # Slow down time itself -- but only in discrete intervals
    # User shouldn't rely on this. The test is here to catch possible
    # backwards-incompatible behavior (which would need a version bump).
    clock = Clock()
    subclock = Subclock(clock)
    subclock.schedule(Animation(subclock, 'speed', 0, time=2))
    clock.advance(1)
    assert subclock.time == 1
    clock.advance(1)
    assert subclock.time == 1.5
    clock.advance(1)
    assert subclock.time == 1.75
    clock.advance(1)
    assert subclock.time == 1.875
    for dummy in range(1000):
        clock.advance(1)
    assert subclock.time == 2



# Encoding: UTF-8

"""Gillcup Actions

Although arbitrary callables can be scheduled on a Gillcup
:class:`~gillcup.Clock`, one frequently schedules objects that are specifically
made for this purpose.
Using :class:`gillcup.Action` allows one to chain actions together in various
ways, allowing the developer to create complex effects.
"""

from __future__ import unicode_literals, division, print_function

import numbers
import functools

from six import callable  # pylint: disable=W0622


class Action(object):
    """A chainable “event” designed for being scheduled.

    As any callable, an Action can be scheduled on a clock, either by
    :meth:`~gillcup.Clock.schedule()`, or by chaining, or, as a shortcut,
    directly from the constructor with the `clock` and `dt` arguments.
    Each Action may only be scheduled *once*.

    Other actions (the lowercase term denotes “arbitrary callables”) may be
    chained to an Action, that is, scheduled to run at some time after the
    Action is run.

    Some Actions may represent a time interval or process rather than a
    discrete point in time. In these cases, chained actions are run after the
    interval is over or the process finishes.

    Actions may be combined to form larger structures using
    :ref:`helper Action subclasses <action-building-blocks>` as building
    blocks.
    As a shorthand, the following operators are available:

        *   ``+`` creates a :class:`~gillcup.actions.Sequence` of actions;
            one is run after the other.
        *   ``|`` creates a :class:`~gillcup.actions.Parallel` construct:
            all actions are started at once.

    The operators can be used with regular callables (which are wrapped in
    :class:`~gillcup.actions.FunctionCaller`), or with numbers
    (which create corresponding :class:`delays <gillcup.actions.Delay>`), or
    with iterables (which get wrapped in :class:`~gillcup.actions.Process`).
    """
    # The states an Animation goes through are:
    # - unscheduled (self.clock is unset)
    # - scheduled
    # - in progress (self.expired == True)
    # - done (self.chain_triggered == True)
    scheduled_time = None

    def __init__(self, clock=None, dt=0):
        super(Action, self).__init__()

        # Set to True once the Action runs
        self.expired = False

        # Set to True once chained actions are scheduled
        self.chain_triggered = False

        # The chained actions
        self._chain = []

        # The clock
        self.clock = None
        if clock:
            clock.schedule(self, dt)
        elif dt != 0:
            # We're not scheduling yet, so dt would be ignored
            raise ValueError('dt specified without a clock')

    def chain(self, action, dt=0):
        """Schedule an action to be scheduled after this Action

        The dt argument can be given to delay the chained action by the
        specified time.

        If this Action has already been called, the chained action is scheduled
        immediately `dt` units after the current time.
        To prevent or modify this behavior, the caller can check the
        :attr:`~gillcup.Action.chain_triggered` attribute.

        Returns the chained action.
        """
        if self.chain_triggered:
            self.clock.schedule(action, dt)
        else:
            self._chain.append((action, dt))
        return action

    @classmethod
    def coerce(cls, value):
        """Coerce value into an action. Called on operands of ``+`` and ``|``.

        Wraps functions in FunctionCallers, and numbers in Delays
        """
        if isinstance(value, Action):
            return value
        elif callable(value):
            return FunctionCaller(value)
        elif isinstance(value, numbers.Real):
            return Delay(value)
        else:
            try:
                iterator = iter(value)
            except TypeError:
                raise ValueError("%s can't be coerced into Action" % value)
            else:
                return Process(iterator)

    def __call__(self):
        """Run this action.

        Subclasses that represent discrete moments in time should call the
        superclass implementation when they are finished running.

        Subclasses that represent time intervals (there's a delay between
        the moment they are called and when they trigger chained actions)
        should call :meth:`~gillcup.Action.expire` when they are called,
        and :meth:`~gillcup.Action.trigger_chain` when they're done.
        """
        self.expire()
        self.trigger_chain()

    def expire(self):
        """Marks the Action as run.

        Subclasses must call this method at the start of
        :meth:`~gillcup.Action.__call__`.
        """
        if self.expired:
            raise RuntimeError('%s was run twice' % self)
        self.expired = True

    def trigger_chain(self):
        """Schedule the chained actions.

        Subclasses must call this method after the Action runs; see
        :meth:`~gillcup.Action.__call__`.
        """
        self.chain_triggered = True
        for dt, chained in self._chain:
            self.clock.schedule(dt, chained)
        self._chain = []

    def schedule_callback(self, clock, time):
        """Called from a clock when this Action is scheduled"""
        if self.clock:
            raise RuntimeError('%s was scheduled twice' % self)
        self.clock = clock
        self.scheduled_time = time

    def __add__(self, other):
        try:
            other = self.coerce(other)
        except ValueError:
            return NotImplemented
        return Sequence(self, other)

    def __radd__(self, other):
        try:
            other = self.coerce(other)
        except ValueError:
            return NotImplemented
        return Sequence(other, self)

    def __or__(self, other):
        try:
            other = self.coerce(other)
        except ValueError:
            return NotImplemented
        return Parallel(self, other)

    def __ror__(self, other):
        try:
            other = self.coerce(other)
        except ValueError:
            return NotImplemented
        return Parallel(other, self)


class ManualAction(Action):
    """An action doesn't end until trigger_chain is called manually
    """
    def __call__(self):
        self.expire()


class FunctionCaller(Action):
    """An Action that calls given `function`, passing `args` and `kwargs` to it

    `function` can be any callable.
    """
    def __init__(self, function, *args, **kwargs):
        super(FunctionCaller, self).__init__()
        self.function = function
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        self.function(*self.args, **self.kwargs)
        super(FunctionCaller, self).__call__()


class Delay(Action):
    """An Action that triggers chained actions after a given delay

    The `kwargs` are passed to :class:`gillcup.Action`'s initializer.
    """
    def __init__(self, time, **kwargs):
        super(Delay, self).__init__(**kwargs)
        self.time = time

    def __call__(self):
        self.expire()
        self.clock.schedule(self.trigger_chain, self.time)


class Sequence(Action):
    """An Action that runs a series of Actions one after the other

    Actions chained to a Sequence are triggered after the last Action in the
    sequence.

    The `kwargs` are passed to :class:`gillcup.Action`'s initializer.
    """
    def __init__(self, *actions, **kwargs):
        super(Sequence, self).__init__(**kwargs)
        self.remaining_actions = list(actions)
        self.remaining_actions.reverse()

    def __call__(self):
        self.expire()
        self._call_next()

    def _call_next(self):
        try:
            action = self.remaining_actions.pop()
        except IndexError:
            self.trigger_chain()
        else:
            action.chain(self._call_next)
            self.clock.schedule(action)

    # XXX: Overload __add__, __radd__?


class Parallel(Action):
    """Starts the given Actions, and triggers chained ones after all are done

    That is, after all the given actions have triggered their chained actions,
    Parallel triggers its own chained actions.

    The `kwargs` are passed to :class:`gillcup.Action`'s initializer.
    """
    def __init__(self, *actions, **kwargs):
        self.remaining_actions = actions
        super(Parallel, self).__init__(**kwargs)

    def __call__(self):
        self.expire()
        for action in self.remaining_actions:
            def _triggered(action=action):
                self._triggered(action)
            action.chain(_triggered)
            self.clock.schedule(action)
        self.remaining_actions = set(self.remaining_actions)

    def _triggered(self, action):
        self.remaining_actions.remove(action)
        if not self.remaining_actions:
            self.trigger_chain()

    # XXX: Overload __or__, __ror__?


class Process(Action):
    """Wraps the given iterable

    When triggered, takes an item from the iterable and schedules it, then
    chains the scheduling of the next item, and so on.
    When the underlying iterator is exhausted, chained actions are run.

    The items in the underlying iterable can be callables, numbers or other
    iterables, as for :class:`~gillcup.actions.Action`'s ``+`` and ``|``
    operators.

    The `kwargs` are passed to :class:`gillcup.Action`'s initializer.

    See :func:`process_generator` for a simple way to create Processes.
    """
    def __init__(self, iterable, **kwargs):
        self.iterator = iter(iterable)
        self.next_action = None
        super(Process, self).__init__(**kwargs)

    def __call__(self):
        self.expire()
        self.do_next()

    def do_next(self):
        """Schedule the next thing from the iterable"""
        try:
            try:
                send = self.iterator.send
            except AttributeError:
                value = next(self.iterator)
            else:
                value = send(self.next_action)
        except StopIteration:
            self.trigger_chain()
        else:
            action = self.next_action = self.coerce(value)
            action.chain(self.do_next)
            self.clock.schedule(action)


def process_generator(func):
    """Decorator for creating :class:`~gillcup.actions.Process`\\ es

    Used as a decorator on a generator function, it allows writing in a
    declarative style instead of callbacks, with ``yield`` statements for
    "asynchronousness".
    """
    @functools.wraps(func)
    def _f(*args, **kwargs):
        return Process(func(*args, **kwargs))
    return _f



# Encoding: UTF-8
"""Gillcup's Animated Properties

To animate Python objects, we need to change values of their attributes over
time.
There are two kinds of changes we can make: *discrete* and *continuous*.
A discrete change happens at a single point in time: for example, an object
is shown, some output is written, a sound starts playing.
:mod:`Actions <gillcup.actions>` are used for effecting
discrete changes.

Continuous changes happen over a period of time: an object smoothly moves
to the left, or a sound fades out.
These changes are made by animating special properties on objects,
using :mod:`Animation <gillcup.animation>` classes on so-called
:mod:`animated properties <gillcup.properties>`.

Under the hood, Gillcup uses Python's `descriptor interface
<http://docs.python.org/howto/descriptor.html>`_ to provide efficient
animated properties.

Assigment to an animated attribute causes the property to get set to the given
value and cancels any running animations on it.
"""

from __future__ import unicode_literals, division, print_function

from gillcup.effect import Effect, ConstantEffect


class AnimatedProperty(object):
    """A scalar animated property

    The idiomatic way to define animated properties is as follows::

        class Tone(object):
            pitch = AnimatedProperty(440)
            volume = AnimatedProperty(0)

    Now, Tone instances will have `pitch` and `volume` set to the corresponding
    defaults, and can be animated.

    The **docstring** argument becomes the property's ``__doc__`` attribute.
    """
    def __init__(self, default, docstring=None):
        self.default = default
        if docstring:
            self.__doc__ = docstring

    def adjust_value(self, values):
        """Convert an animation's ``*args`` values into a property value

        For scalar properties, this converts a 1-tuple into its only element
        """
        [value] = values
        return value

    def get_target_property(self):
        """Return a property used for a dynamic animation's target
        """
        # Since AnimatedProperty doesn't care about its name, we can just
        # reuse it
        return self

    def __get__(self, instance, owner):
        if instance:
            return self.get_effect(instance).value
        else:
            return self

    def __set__(self, instance, value):
        self.animate(instance, ConstantEffect(value))
        self.do_replacements(instance)

    def __delete__(self, instance):
        self.animate(instance, ConstantEffect(self.default))
        self.do_replacements(instance)

    def get_effect(self, instance):
        """Get the current effect; possibly create a default one beforehand"""
        # pylint: disable=W0212
        try:
            effects = instance.__gillcup_effects
        except AttributeError:
            effects = instance.__gillcup_effects = {}
        try:
            effect = effects[self]
        except KeyError:
            effect = effects[self] = ConstantEffect(self.default)
        return effect

    def animate(self, instance, animation):
        """Set a new effect on this property; return the old one"""
        # pylint: disable=W0212
        parent = self.get_effect(instance)
        instance.__gillcup_effects[self] = animation
        return parent.get_replacement()

    def do_replacements(self, instance):
        """Possibly replace current effect w/ a more lightweight equivalent"""
        # pylint: disable=W0212
        try:
            effects = instance.__gillcup_effects
            current_effect = effects[self]
        except (AttributeError, KeyError):
            pass
        else:
            effects[self] = current_effect.get_replacement()

    def tween_values(self, function, parent_value, value):
        """Call a scalar tween function on two values.
        """
        return function(parent_value, value)


class TupleProperty(AnimatedProperty):
    """A tuple animated property

    Iterating the TupleProperty itself yields sub-properties that can be
    animated individually.
    The intended idiom for declaring a tuple property is::

        x, y, z = position = TupleProperty(0, 0, 0)
    """
    def __init__(self, *default, **kwargs):
        super(TupleProperty, self).__init__(default, **kwargs)
        self.size = len(default)
        self.subproperties = [
            _TupleElementProperty(self, i) for i in range(self.size)]

    def adjust_value(self, value):
        """Convert an animation's ``*args`` values into a property value

        For tuple properties, return the tuple unchanged
        """
        return value

    def __set__(self, instance, value):
        self.animate(instance, ConstantEffect(self.adjust_value(value)))

    def __iter__(self):
        return iter(self.subproperties)

    def tween_values(self, function, parent_value, value):
        """Call a scalar tween function on two values.

        Calls the function on corresponding pairs of elements, returns
        the tuple of results
        """
        return tuple(map(function, parent_value, value))


class _TupleElementProperty(AnimatedProperty):
    """Animated property for one element of a TupleProperty
    """
    def __init__(self, parent, index):
        super(_TupleElementProperty, self).__init__(parent.default[index])
        self.parent = parent
        self.index = index

    def get_effect(self, instance):
        parent_effect = self.parent.get_effect(instance)
        return _TupleExtractEffect(parent_effect, self.index)

    def animate(self, instance, animation):
        tuple_effect = _TupleMakeEffect(animation, self.index)
        parent = self.parent.animate(instance, tuple_effect)
        tuple_effect.parent = parent
        return _TupleExtractEffect(parent, self.index)

    def do_replacements(self, instance):
        self.parent.do_replacements(instance)


class _TupleExtractEffect(Effect):
    """Effect that extracts one element of a tuple
    """
    def __init__(self, parent, index):
        super(_TupleExtractEffect, self).__init__()
        self.parent = parent
        self.index = index

    @property
    def value(self):
        """Value to be used for the property this effect is on"""
        return self.parent.value[self.index]

    def get_replacement(self):
        self.parent = self.parent.get_replacement()
        if (isinstance(self.parent, _TupleMakeEffect) and
                self.parent.index == self.index):
            return self.parent.element_effect
        elif isinstance(self.parent, ConstantEffect):
            return ConstantEffect(self.value)
        else:
            return self


class _TupleMakeEffect(Effect):
    """Effect that recombines one changed element of a tuple with the rest

    `element_effect` is an Effect whose `value` is used for the changed element

    The `parent` attribute has the Effect with the original, full tuple.
    This attribute must be set after instantiation.
    """
    element_effect = parent = None

    def __init__(self, element_effect, index):
        super(_TupleMakeEffect, self).__init__()
        self.element_effect = element_effect
        self.index = index

    @property
    def value(self):
        """Value to be used for the property this effect is on"""
        element_effect = self.element_effect
        return tuple(
            element_effect.value if i == self.index else val
            for i, val in enumerate(self.parent.value))

    def get_replacement(self):
        self.parent = self.parent.get_replacement()
        self.element_effect = self.element_effect.get_replacement()
        if (isinstance(self.parent, _TupleMakeEffect) and
                self.parent.index == self.index):
            self.parent = self.parent.parent
            return self.get_replacement()
        elif (isinstance(self.parent, ConstantEffect) and
                isinstance(self.element_effect, ConstantEffect)):
            return ConstantEffect(self.value)
        else:
            return self


class ScaleProperty(TupleProperty):
    """A TupleProperty used for scales or sizes in multiple dimensions

    It acts as a regular TupleProperty, but supports scalars or short tuples in
    assignment or animation.

    Instead of a default value, __init__ takes the number of dimensions;
    the default value will be ``(1,) * num_dimensions``.

    If a scalar, or a tuple with only one element, is given, the value is
    repeated across all dimensions.
    If another short tuple is given, the remaining dimensions are set to 1.

    For example, given::

        width, height, length = size = ScaleProperty(3)

    the following pairs are equivalent::

        obj.size = 2
        obj.size = 2, 2, 2

        obj.size = 2, 3
        obj.size = 2, 3, 1

        obj.size = 2,
        obj.size = 2, 2, 2

    Similarly, ``Animation(obj, 'size', 2)`` is equivalent to
    ``Animation(obj, 'size', 2, 2, 2)``.
    """
    def __init__(self, num_dimensions, **kwargs):
        super(ScaleProperty, self).__init__(*(1, ) * num_dimensions, **kwargs)

    def adjust_value(self, value):
        """Expand the given tuple or scalar to a tuple of len=num_dimensions
        """
        try:
            size = len(value)
        except TypeError:
            return (value, ) * self.size
        if size == self.size:
            return value
        elif size == 1:
            return value * self.size
        elif size < self.size:
            return value + (1, ) * (self.size - size)
        else:
            raise ValueError('Too many dimensions for ScaleProperty')


class VectorProperty(TupleProperty):
    """A TupleProperty used for vectors

    It acts as a regular TupleProperty, but supports short tuples in
    assignment or animation by setting all remaining dimensions to 0.

    Instead of a default value, __init__ takes the number of dimensions;
    the default value will be ``(0,) * num_dimensions``.

    For example, given::

        x, y, z = position = VectorProperty(3)

    the following pairs are equivalent::

        obj.position = 2, 3
        obj.position = 2, 3, 0

        obj.position = 2,
        obj.position = 2, 0, 0

    Similarly, ``Animation(obj, 'position', 1, 2)`` is equivalent to
    ``Animation(obj, 'position', 1, 2, 0)``.
    """
    def __init__(self, num_dimensions, **kwargs):
        super(VectorProperty, self).__init__(*(0, ) * num_dimensions, **kwargs)

    def adjust_value(self, value):
        """Expand the given tuple to the correct number of dimensions
        """
        size = len(value)
        if size == self.size:
            return value
        elif size < self.size:
            return value + (0, ) * (self.size - size)
        else:
            raise ValueError('Too many dimensions for VectorProperty')



# Encoding: UTF-8
"""Gillcup's Clock Class

In Gillcup, animation means two things: running code at specified times,
and changing object properties with time.

You will notice that the preceding sentence mentions time quite a lot. But
what is this time?

You could determine time by looking at the computer's clock, but that would
only work with real-time animations. When you'd want to render a movie,
where each frame takes 2 seconds to draw and there are 25 frames per
second, you'd be stuck.
That's why Gillcup introduces a flexible source of time, the Clock, which
keeps track of time and schedules actions.

Time is measured in “time units”.
What a time unit means is entirely up to the application – it could be
seconds, movie/simulation frames, etc.
"""

from __future__ import unicode_literals, division, print_function

import collections
import weakref
import heapq

from gillcup.properties import AnimatedProperty

_HeapEntry = collections.namedtuple('EventHeapEntry', 'time index action')

# Next action index; used to keep FIFO ordering for actions scheduled
# for the same time
next_index = 0


class Clock(object):
    """Keeps track of time and schedules events.

    Attributes:

        .. attribute:: time

            The current time on the clock. Never assign to it directly;
            use :meth:`~gillcup.Clock.advance()` instead.
    """
    def __init__(self):
        # Time on the clock
        self.time = 0

        # Heap queue of scheduled actions
        self.events = []

        # Update functions (see `schedule_update_function`)
        self.update_functions = WeakSet()

        # Recursion guard flag for advance()
        self.advancing = False

        # List of dependent clocks
        self._subclocks = set()

    speed = AnimatedProperty(1, docstring="""Speed of the clock.

    When calling update(), the interval is multiplied by this value.

    The speed is an AnimatedProperty. When changing, beware that it is only
    checked when advance() is called or when a scheduled action is run,
    so speed animations will be only approximate.
    For better accuracy, call :meth:`~gillcup.Clock.advance`
    with small *dt*, or schedule a periodic dummy action at small inervals.
    """)

    @property
    def _next_event(self):
        try:
            event = self.events[0]
            events = [(event.time - self.time, event.index, self, event)]
        except IndexError:
            events = []
        for subclock in self._subclocks:
            event = subclock._next_event  # pylint: disable=W0212
            if event:
                remain, index, clock, event = event
                try:
                    remain /= subclock.speed
                except ZeroDivisionError:
                    # zero speed – events never happen
                    pass
                else:
                    events.append((remain, index, clock, event))
        try:
            return min(events)
        except ValueError:
            return None

    def advance(self, dt):
        """Call to advance the clock's time

        Steps the clock dt units to the future, pausing at times when actions
        are scheduled, and running them.

        Attempting to move to the past (dt<0) will raise an error.
        """
        dt *= self.speed
        if dt < 0:
            raise ValueError('Moving backwards in time')
        if self.advancing:
            raise RuntimeError('Clock.advance called recursively')
        self.advancing = True
        try:
            while True:
                event = self._next_event
                if not event:
                    break
                event_dt, _index, clock, event = event
                if event_dt > dt:
                    break
                if dt:
                    self._advance(event_dt)
                    self._run_update_functions()
                dt -= event_dt
                _evt = heapq.heappop(clock.events)
                assert _evt is event and clock.time == event.time
                clock.time = event.time
                event.action()
            if dt:
                self._advance(dt)
                self._run_update_functions()
        finally:
            self.advancing = False

    def _advance(self, dt):
        self.time += dt
        for subclock in self._subclocks:
            subclock._advance(dt * subclock.speed)  # pylint: disable=W0212

    def schedule(self, action, dt=0):
        """Schedule an action to be run "dt" time units from the current time

        Scheduling is stable:  if two things are scheduled for the same
        time, they will be called in the order they were scheduled.

        Scheduling an action in the past (dt<0) will raise an error.

        If the scheduled callable has a “schedule_callback” method, it will
        be called with the clock and the time it's been scheduled at.
        """
        global next_index
        if dt < 0:
            raise ValueError('Scheduling an action in the past')
        next_index += 1
        scheduled_time = self.time + dt
        entry = _HeapEntry(scheduled_time, next_index, action)
        heapq.heappush(self.events, entry)
        try:
            schedule_callback = action.schedule_callback
        except AttributeError:
            pass
        else:
            schedule_callback(self, scheduled_time)

    def schedule_update_function(self, function):
        """Schedule a function to be called every time the clock advances

        Then function will be called a lot, so it shouldn't be very expensive.

        Only a weak reference is made to the function, so the caller should
        ensure another reference to it is retained as long as it should be
        called.
        """
        self.update_functions.add(function)

    def unschedule_update_function(self, function):
        """Unschedule a function scheduled by `schedule_update_function`
        """
        # Don't raise an error if the function is no longer there – we're
        # dealing with weakrefs.
        self.update_functions.discard(function)

    def _run_update_functions(self):
        for update_function in list(self.update_functions):
            update_function()
        for subclock in self._subclocks:
            subclock._run_update_functions()  # pylint: disable=W0212


class Subclock(Clock):
    """A Clock that advances in sync with another Clock

    A Subclock advances whenever its *parent* clock does.
    Its `speed` attribute specifies the relative speed relative to the parent
    clock. For example, if speed==2, the subclock will run twice as fast as its
    parent clock.

    Unlike clocks synchronized via actions or update functions, the actions
    scheduled on a parent Clock and all subclocks are run in the correct
    sequence, with all clocks at the correct times when each action is run.
    """

    def __init__(self, parent, speed=1):
        super(Subclock, self).__init__()
        self.speed = speed
        parent._subclocks.add(self)  # pylint: disable=W0212


try:  # pragma: no cover
    WeakSet = weakref.WeakSet  # pylint: disable=E1101
except AttributeError:  # pragma: no cover

    class WeakSet(weakref.WeakKeyDictionary):
        """Stripped-down WeakSet implementation for Python 2.6

        (only defines the methods we need)
        """
        def add(self, item):
            """Add an item to the set"""
            self[item] = None

        def discard(self, item):
            """Remove an item from the set"""
            self.pop(item)



#! /usr/bin/python
# Encoding: UTF-8

"""
Adapded roughly and partially from Robert Penner's Easing Equations, as they
appear in the Qt library. The original license follows:

TERMS OF USE - EASING EQUATIONS

Open source under the BSD License.

Copyright © 2001 Robert Penner

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   * Neither the name of the author nor the names of contributors may be
     used to endorse or promote products derived from this software without
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import unicode_literals, division, print_function

import sys
import functools
import math


def normalized(func):
    """Decorator that normalizes an easing function

    Normalizing is done so that func(0) == 0 and func(1) == 1.
    """
    minimum = func(0)
    maximum = func(1)
    if (minimum, maximum) == (0, 1):
        return func
    range_ = maximum - minimum

    @functools.wraps(func)
    def _normalized(t):
        return (func(t) - minimum) / range_
    _add_postfix(_normalized, func, 'normalized')

    return _normalized


def _add_postfix(decorated, func, postfix):
    try:
        decorated.__name__ = str(func.__name__ + '_' + postfix)
    except AttributeError:  # pragma: no cover
        pass


def ease_out(func):
    """Given an "in" easing function, return corresponding "out" function"""
    def _ease_out(t):
        return 1 - func(1 - t)
    _add_postfix(_ease_out, func, 'out')
    return _ease_out


def ease_out_in(func):
    """Given an "in" easing function, return corresponding "out-in" function"""
    def _ease_out_in(t):
        if t < 0.5:
            return (1 - func(1 - 2 * t)) / 2
        else:
            return func(2 * (t - .5)) / 2 + .5
    _add_postfix(_ease_out_in, func, 'out_in')
    return _ease_out_in


def ease_in_out(func):
    """Given an "in" easing function, return corresponding "in-out" function"""
    def _ease_in_out(t):
        if t < 0.5:
            return func(2 * t) / 2
        else:
            return 1 - func(1 - 2 * (t - .5)) / 2
    _add_postfix(_ease_in_out, func, 'in_out')
    return _ease_in_out


def ease_in(func):
    """Return ``func`` itself. Included for symmetry."""
    return func


def easefunc(func):
    """Decorator for easing functions.

    Adds the **in_**, **out**, **in_out** and **out_in** attributes to
    an easing function.
    """
    func.in_ = ease_in(func)
    func.out = ease_out(func)
    func.in_out = ease_in_out(func)
    func.out_in = ease_out_in(func)
    return func


@easefunc
def linear(t):
    """Linear interpolation

    t → t
    """
    return t


@easefunc
def quadratic(t):
    """Quadratic easing

    t → t**2
    """
    return t ** 2


@easefunc
def cubic(t):
    """Cubic easing

    t → t**3
    """
    return t ** 3


@easefunc
def quartic(t):
    """Quartic easing

    t → t**4
    """
    return t ** 4


@easefunc
def quintic(t):
    """Quintic easing

    t → t**5
    """
    return t ** 5


@easefunc
def sine(t):
    """Sinusoidal easing

    Quarter of a cosine wave
    """
    return (-math.cos(t / 2 * math.pi) + 1)


@easefunc
def exponential(t):
    """Exponential easing
    """
    if t in (0, 1):
        return t
    else:
        return 2.0 ** (10 * (t - 1)) - 0.001


@easefunc
def circular(t):
    """Circular easing
    """
    try:
        return 1 - math.sqrt(1 - t * t)
    except ValueError:
        return 1


def elastic(period, amplitude=1):
    """Elastic easing factory
    """
    @easefunc
    def _elastic(t):
        b = exponential(t) * math.cos((1 - t) * 2 * math.pi / period)
        return b * amplitude
    return _elastic


def overshoot(amount):
    """Overshoot easing factory
    """
    @easefunc
    def _overshoot(t):
        return t * t * ((amount + 1) * t - amount)
    return _overshoot


def _bounce_helper(t, c, a):
    t = 1 - t
    if t == 1:
        return 1 - c
    if t < 4 / 11:
        return 1 - c * (7.5625 * t * t)
    elif t < 8 / 11:
        t -= 6 / 11.0
        return 1 + a * (1. - (7.5625 * t * t + .75)) - c
    elif t < 10 / 11:
        t -= 9 / 11
        return 1 + a * (1. - (7.5625 * t * t + .9375)) - c
    else:
        t -= 21 / 22
        return 1 + a * (1. - (7.5625 * t * t + .984375)) - c


def bounce(amplitude):
    """Bounce easing factory
    """
    @easefunc
    def _bounce(t):
        return _bounce_helper(t, 1, amplitude)
    return _bounce


# Execute the file for a gallery of easing funcs (matplotlib must be installed)

elastic_example = elastic(.15)
overshoot_example = overshoot(1)
bounce_example = bounce(1)


def showcase(
        items='(poly) sine exponential circular elastic_example '
        'overshoot_example bounce_example'.split(),
        filename=None,
):  # pragma: no cover; pylint: disable=R0914, W0621, W0404, F0401
    """Show graphs of the easing functions in this module
    """
    import pylab
    from matplotlib import pyplot
    from matplotlib.ticker import MultipleLocator

    pyplot.figure(figsize=(7, 7))
    pyplot.subplots_adjust(wspace=0.3, hspace=0.7)

    time = pylab.arange(0.0, 1.01, 0.01)
    for i, n in enumerate(items):
        if n == '(poly)':
            funcs = 'linear quadratic cubic quartic quintic'.split()
        else:
            funcs = [n]
        for j, a in enumerate('in_ out in_out out_in'.split()):
            ax = pylab.subplot(len(items), 4, i * 4 + 1 + j)
            ax.xaxis.set_major_locator(MultipleLocator(1))
            ax.yaxis.set_major_locator(MultipleLocator(1))
            for funcname in funcs:
                func = getattr(globals()[funcname], a)
                s = [func(t) for t in time]
                pylab.plot(time, s, linewidth=1.0)
                if i == 0:
                    pylab.title('.' + a)
                if j == 0:
                    title, sep, _end = n.partition('_example')
                    pylab.ylabel(title)
    if filename:
        pylab.savefig(filename, transparent=True)
    else:
        pylab.show()


if __name__ == '__main__':
    try:
        filename = sys.argv[1]
    except IndexError:
        filename = None
    showcase(filename=filename)



"""Provide a run() function that wraps the pytest runner in a unittest suite.

The goal is to:
- have `setup.py test` work
- have `tox` work, even with Distribute's use_2to3 option
- have `py.test` work
- not include a binary blob in the repository (which pytest actually recommends
    at http://pytest.org/latest/goodpractises.html)

These requirements are, alas, rather difficult to juggle. I would appreciate
a better solution than the following hack.
"""

import unittest
import os
import sys

import six

skip_lint = six.PY3 or hasattr(sys, 'pypy_version_info')
if not skip_lint:
    from pylint import lint

import pytest

skipIf = getattr(unittest, 'skipIf', None)
if skipIf is None:  # pragma: no cover
    def skipIf(condition, reason):  # pylint: disable=E0102
        """Replacement for unittest.skipIf for pythons that don't have it"""
        def _skip(func):
            if condition:
                print('skipping test: %s' % reason)
                return(reason)
            else:
                return func
        return _skip


def test_suite(basefile=__file__):  # pragma: no cover
    """Top-level test function"""
    test_dir = os.path.abspath(os.path.join(os.path.dirname(basefile), '..'))

    class PytestWrapper(unittest.TestCase):
        """TestCase that wraps pytest & pylint"""
        def test_wraper(self):
            """Run pytest"""
            errno = pytest.main([os.path.dirname(basefile), '-v'])
            assert not errno, 'pytest failed'

        @skipIf(skip_lint, 'pylint not supported on this python')
        def lint_wraper(self):
            """Run pylint"""
            rc_path = os.path.join(test_dir, '..', '.pylintrc')
            try:
                print([test_dir, '--rcfile', rc_path])
                print(basefile)
                errno = lint.Run([test_dir, '--rcfile', rc_path])
            except SystemExit:
                errno = sys.exc_info()[1].code
            assert not errno, 'pylint failed'

    return unittest.TestSuite([
        PytestWrapper('test_wraper'),
        PytestWrapper('lint_wraper')])



"""Gillcup, a Python animation library

Gillcup provides a number of modules:

.. toctree::
    :maxdepth: 1

    clock
    actions
    properties
    animation
    effect
    easing

The most interesting classes of each module are exported directly
from the gillcup package:

* :class:`~gillcup.Clock` (from :mod:`gillcup.clock`)
* :class:`~gillcup.Subclock` (from :mod:`gillcup.clock`)
* :class:`~gillcup.Action` (from :mod:`gillcup.actions`)
* :class:`~gillcup.AnimatedProperty` (from :mod:`gillcup.properties`)
* :class:`~gillcup.TupleProperty` (from :mod:`gillcup.properties`)
* :class:`~gillcup.Animation` (from :mod:`gillcup.animation`)
* :class:`~gillcup.Effect` (from :mod:`gillcup.effect`)
* :class:`~gillcup.ConstantEffect` (from :mod:`gillcup.effect`)
"""

__version__ = '0.2.0-beta.2'
__version_info__ = (0, 2, 0, 'beta', 2)

from gillcup.clock import Clock, Subclock
from gillcup.actions import Action
from gillcup.properties import AnimatedProperty, TupleProperty
from gillcup.animation import Animation
from gillcup.effect import Effect, ConstantEffect



"""Effect base & helper classes

The Effect is the base class that modify an AnimatedProperty.
:class:`~gillcup.Animation` is Effect's most important subclass.

Each Effect can be applied to one or more properties on one or more objects.
The value of these properties is then provided by the Effect's ``value``
property.
"""


class Effect(object):
    """Object that changes an AnimatedProperty

    Effects should have a `value` attribute that provides a value for the
    property.
    """
    is_constant = False

    def get_replacement(self):
        """Return an equivalent effect

        When it's sure that the effect's value won't change any more, this
        method can return a :class:`~gillcup.ConstantEffect` to free resources.
        """
        return self

    def apply_to(self, instance, property_name):
        """Apply this effect to an ``instance``'s AnimatedProperty"""
        getattr(type(instance), property_name).animate(instance, self)


class ConstantEffect(Effect):
    """An Effect that provides a constant value"""
    is_constant = True

    def __init__(self, value):
        super(ConstantEffect, self).__init__()
        self.value = value




import sys

from setuptools import setup, find_packages

setup_args = dict(
    name='gillcup',
    version='0.2.0-beta.2',  # XXX: Duplicated in __init__.py
    packages=find_packages(),

    description="""An animation framework for Python""",
    author='Petr Viktorin',
    author_email='encukou@gmail.com',
    url='http://pypi.python.org/pypi/gillcup/',
    install_requires=['six>=1.1'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Topic :: Software Development :: Libraries',
    ],

    tests_require=['pytest>=2.2', 'pytest-pep8'],
    package_data={'': ['.pylintrc']},
    test_suite='gillcup.test.test_suite',
)

if sys.version_info < (3, 0):
    setup_args['tests_require'].append('pylint')

if __name__ == '__main__':
    setup(**setup_args)



"""Tests for the gillcup.action module"""

from pytest import raises, mark

from gillcup import Clock, Action, actions


class TimeAppendingAction(Action):
    """Action that appends the current time to a list when triggered"""
    def __init__(self, lst, *args, **kwargs):
        super(TimeAppendingAction, self).__init__(*args, **kwargs)
        self.list = lst

    def __call__(self):
        self.list.append(self.clock.time)
        super(TimeAppendingAction, self).__call__()


def test_scheduling():
    """Test basic scheduling of actions"""
    clock = Clock()
    lst = []
    action = TimeAppendingAction(lst)
    clock.schedule(action, 1)
    clock.advance(1)
    assert lst == [1]
    TimeAppendingAction(lst, clock, 1)
    clock.advance(1)
    assert lst == [1, 2]
    with raises(ValueError):
        TimeAppendingAction(lst, dt=1)


def test_double_schedule():
    """Test scheduling an Action twice is blocked"""
    clock = Clock()
    lst = []
    action = TimeAppendingAction(lst, clock, 1)
    with raises(RuntimeError):
        clock.schedule(action, 1)


def test_double_run():
    """Test running an Action twice is blocked"""
    clock = Clock()
    lst = []
    action = TimeAppendingAction(lst, clock, 1)
    clock.advance(1)
    with raises(RuntimeError):
        action()


def test_simple_chaining():
    """Test basic chaining"""
    clock = Clock()
    lst = []
    action = TimeAppendingAction(lst, clock, 1)
    action.chain(TimeAppendingAction(lst), 1)
    clock.advance(1)
    assert lst == [1]
    clock.advance(1)
    assert lst == [1, 2]


def test_delayed_chaining():
    """Test chaining with a delay"""
    clock = Clock()
    lst = []
    action = TimeAppendingAction(lst, clock, 1)
    clock.advance(1)
    assert lst == [1]
    action.chain(TimeAppendingAction(lst), 1)
    assert lst == [1]
    clock.advance(1)
    assert lst == [1, 2]
    action.chain(TimeAppendingAction(lst), 1)
    assert lst == [1, 2]
    clock.advance(1)
    assert lst == [1, 2, 3]


def test_function_caller():
    """Test the FunctionCaller action"""
    clock = Clock()
    lst = []
    action = TimeAppendingAction(lst, clock, 1)
    action = action.chain(actions.FunctionCaller(lst.append, 'a'), 1)
    action = action.chain(TimeAppendingAction(lst), 1)  # pylint: disable=E1101
    clock.advance(1)
    assert lst == [1]
    clock.advance(1)
    assert lst == [1, 'a']
    clock.advance(1)
    assert lst == [1, 'a', 3]


def test_delay():
    """Test the Delay action"""
    clock = Clock()
    lst = []
    action = TimeAppendingAction(lst, clock, 1)
    action = action.chain(actions.Delay(1))
    action = action.chain(TimeAppendingAction(lst))
    clock.advance(1)
    assert lst == [1]
    clock.advance(1)
    assert lst == [1, 2]


def test_sequence():
    """Test the Sequence action (explicit instantiation)"""
    clock = Clock()
    lst = []
    action = actions.Sequence(
        TimeAppendingAction(lst),
        actions.Delay(1),
        TimeAppendingAction(lst),
        actions.Delay(1),
        TimeAppendingAction(lst))
    action.chain(TimeAppendingAction(lst))
    clock.schedule(action)
    clock.advance(2)
    assert lst == [0, 1, 2, 2]


def test_parallel():
    """Test the Parallel action (explicit instantiation)"""
    clock = Clock()
    lst = []
    action = actions.Parallel(
        TimeAppendingAction(lst),
        TimeAppendingAction(lst),
        actions.Sequence(actions.Delay(1), TimeAppendingAction(lst)),
        actions.Sequence(actions.Delay(2), TimeAppendingAction(lst)))
    action.chain(TimeAppendingAction(lst))
    clock.schedule(action)
    clock.advance(2)
    assert lst == [0, 0, 1, 2, 2]


def test_sequence_operator():
    """Test the Sequence action (via + operator)"""
    clock = Clock()
    lst = []
    action = TimeAppendingAction(lst) + 1 + TimeAppendingAction(lst)
    action = 1 + action + 1 + (lambda: lst.append('a'))
    action += lambda: lst.append('b')
    action.chain(lambda: lst.append('end'))
    clock.schedule(action)
    assert lst == []
    clock.advance(2)
    assert lst == [1, 2]
    clock.advance(2)
    assert lst == [1, 2, 'a', 'b', 'end']
    with raises(TypeError):
        action += None
    with raises(TypeError):
        action = None + action


def test_parallel_operator():
    """Test the Parallel action (via | operator)"""
    clock = Clock()
    lst = []
    action = 0.5 + TimeAppendingAction(lst) | 1 + TimeAppendingAction(lst)
    action = 2 + TimeAppendingAction(lst) | action | (lambda: lst.append('a'))
    action |= 3 + TimeAppendingAction(lst)
    action.chain(lambda: lst.append('end'))
    action = 23 | action
    action.chain(lambda: lst.append('real end'))
    clock.schedule(action)
    assert lst == []
    clock.advance(3)
    assert lst == ['a', 0.5, 1, 2, 3, 'end']
    clock.advance(20)
    assert lst == ['a', 0.5, 1, 2, 3, 'end', 'real end']
    with raises(TypeError):
        action |= None
    with raises(TypeError):
        action = None | action


def test_chain_on_expired():
    """Test that stuff chained on an expired action is scheduled"""
    clock = Clock()
    lst = []
    action = actions.Delay(3)
    clock.schedule(action)
    action.chain(lambda: lst.append('a'), 1)
    clock.advance(1)
    assert lst == []
    action.chain(lambda: lst.append('b'), 1)
    clock.advance(1)
    assert lst == []
    clock.advance(1)
    assert lst == []
    clock.advance(1)
    assert lst == ['a', 'b']
    action.chain(lambda: lst.append('c'), 1)
    clock.advance(0)
    assert lst == ['a', 'b']
    clock.advance(1)
    assert lst == ['a', 'b', 'c']


@mark.parametrize(('maker'), [
    lambda g: Action.coerce(g()),
    lambda g: actions.Process(g()),
    lambda g: actions.process_generator(g)()])
def test_generator(maker):
    """Test actions.Generator"""
    clock = Clock()
    lst = []

    def _generator():
        yield 1
        lst.append('a')
        yield [1, TimeAppendingAction(lst)]
        lst.append('b')
        yielded = yield TimeAppendingAction(lst)
        yielded.chain(lambda: lst.append('c'))
        yield 1

    action = maker(_generator)
    clock.schedule(action)
    action.chain(TimeAppendingAction(lst))
    assert lst == []
    clock.advance(1)
    assert lst == ['a']
    clock.advance(1)
    assert lst == ['a', 2, 'b', 2, 'c']
    clock.advance(1)
    assert lst == ['a', 2, 'b', 2, 'c', 3]
